
#ifndef SUBSURFACE_H
#define SUBSURFACE_H

#include "pbrt.h"
#include "primitive.h"
#include "geometry.h"
#include "shape.h"
#include "color.h"

#include <vector>
#include <map>

//forward decl
class SubSurface;

/**
 * class used to store irradiance sample
 */
class ESample {
  public:
    ESample(Spectrum& e, float a, const Point& pt, SubSurface* ss):
        Ep(e), Ap(a), p(pt), subSurf(ss) {}
        
    ESample():
        Ep(), Ap(0), p(0,0,0), subSurf(NULL) {}
    
    /** get bounding box for sample */
    BBox getBBox();    
    
    /** calculate radiant exitance */
    Spectrum Mo(const Point &p, const Normal &n, const Vector &wo);
    
    /** position weight */
    float getPositionWeight(int color);
    
    Spectrum Ep; // irradiance
    float Ap; // area
    Point p; // location
    SubSurface* subSurf; // pointer to obj
};

//print
inline ostream &operator<<(ostream &os, const ESample &v) {
    os << "p: " << v.p << " Ap: " << v.Ap << " Ep: " << v.Ep << " ss: " << v.subSurf;
    return os;
}


/**
 *  Octree node
 */
struct SSOctNode {
    SSOctNode(): depth(0) {
        for (int i = 0; i < 8; ++i)
            children[i] = NULL;
    }
    ~SSOctNode() {
        for (int i = 0; i < 8; ++i)
            delete children[i];
    }
    
    /** summarize children */
    void summarize(SubSurface* ss, int color);
    
    /** get solid angle spanned by points */
    float deltaW(const Point& x);
    
    SSOctNode *children[8];
    vector<ESample> data;
    ESample summary;
    BBox bound;
    int depth;
    
    friend class SubSurface;
};


/**
 *  Modified copy of octree.h for subsurface 
 */
class SSOctree {
public:
    // Octree Public Methods
    SSOctree(int md = 16)
        : maxDepth(md) {}
    
    /** add sample */
    void Add(ESample dataItem, const BBox dataBound) {
        addPrivate(&root, bound, dataItem, dataBound,
            DistanceSquared(dataBound.pMin, dataBound.pMax));
    }
    
    /* debug */
    void print();
    
    /* debug */
    void print(SSOctNode* node, int depth);
    
    /* whether tree is empty */
    bool empty();

private:
    // Octree Private Methods
    void addPrivate(SSOctNode *node, const BBox nodeBound,
                    ESample dataItem, const BBox dataBound, float diag2,
                    int depth = 0);
            
    // Octree Private Data
    int maxDepth;
    BBox bound;
    SSOctNode root;
    
    friend class SubSurface;
};


/**
 *
 * Class used to calculate diffuse reflection from subsurface scattering effects.
 * Given a set of primitives, and irradiance at samples on the primitives,
 * calculates outgoing radiance from the dipole approximation.  One instance of
 * this class is associated with a set of primitives with the same subsurfid,
 * and handles calculations for any point on those primitives
 * 
 */
class SubSurface {
  public:
  
    /** constructor */
    SubSurface();
  
    /** calculate radiant exitance */
    Spectrum Lo(const Point &p, const Normal &n, const Vector &wo);
  
    /** traverse tree to find Mo */
    Spectrum traverseMo(SSOctNode* node, const Point &p, 
                        const Normal &n, const Vector &wo,
                        vector<ESample>& samps);
                         
    /** add a primitive */
    void addPrim(Reference<Primitive> p);
    
    /** set scattering params */
    void setParams(vector<float>& vals);

    /** get the list of primitives */
    vector< Reference<Primitive> >& getPrims() { return prims; }

    /** add an irradiance sample */
    void addSample(Spectrum& e, float area, const Point& loc, int color);
    
    /** summarize the trees */
    void summarize();
    
    /** holds refs to primitives of this subsurface object */
    vector< Reference<Primitive> > prims;
    float area;
    Spectrum sigma_sp;
    Spectrum sigma_a;
    float eta;
    
    // calculated constants
    float F_dr;
    float F_dt;
    float A;
    Spectrum sigma_tp;
    Spectrum sigma_tr;
    Spectrum z_r;
    Spectrum z_v;
    Spectrum lu;
    bool summarized;
    float sampleFactor;
    
    //E samples
    SSOctree samples[3];
};

/** get the global map of subsurface objects */
map<int, SubSurface>& getSubSurfaceMap();

/** method to add a primitive to a subsurface obj according to id */
void addSubSurfacePrimitive(Reference<Primitive> p);

#endif // SUBSURFACE_H
