
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// subsurfacescattering.cpp*
#include "pbrt.h"
#include "transport.h"
#include "scene.h"
#include "mc.h"
#include "kdtree.h"
#include "octree.h"
#include "sampling.h"
#include "subsurface.h"
#include <iostream>
using namespace std;

// SubsurfaceIntegrator Local Declarations
struct Photon;
struct ClosePhoton;
struct PhotonProcess;
struct IrradianceSample;
struct IrradProcess;
class SubsurfaceIntegrator : public SurfaceIntegrator {
public:
    // SubsurfaceIntegrator Public Methods
    SubsurfaceIntegrator(int ncaus, int ndir, int nindir, int nLookup, int ns, int mdepth,
        float maxdist, bool finalGather, int gatherSamples, int irradianceSamples, float maxError,
        bool useIrradianceCache);
    ~SubsurfaceIntegrator();
    Spectrum Li(const Scene *scene, const RayDifferential &ray,
        const Sample *sample, float *alpha) const;
    void RequestSamples(Sample *sample, const Scene *scene);
    void Preprocess(const Scene *);
    Spectrum computeIrradiance(const Scene *scene, const Point &p, const Normal &n) const;

private:
    // SubsurfaceIntegrator Private Methods
    static inline bool unsuccessful(int needed, int found, int shot) {
        return (found < needed &&
                (found == 0 || found < shot / 1024));
    }
    // Initialize photon maps
    void initPhotonMaps(const Scene *scene);
    // Initialize photon irradiance values
    void initPhotonIrradiance(const Scene *scene);
    // Trace photons through scene and add to photon maps
    void tracePhotons(const Scene *scene, RayDifferential &ray,
        Spectrum &alpha, bool *causticDone, bool *directDone, bool *indirectDone,
        StatsCounter &nshot, ProgressReporter &progress);
    // Interpolate irradiance from irradiance cache
    bool interpolateIrradiance(const Scene *scene, const Point &p, const Normal &n, Spectrum *E) const;
    // Add irradiance sample to cache
    void addIrradianceSample(const Scene *scene, const Spectrum &E, const Point &p,
        const Normal &n, const float dist) const;
    // Compute photon irradiance
    Spectrum computePhotonIrradiance(const Scene *scene, const Point &p, const Normal &n) const;
    // Return radiance at photon
    static Spectrum LPhoton(KdTree<Photon, PhotonProcess> *map,
        int nPaths, int nLookup, BSDF *bsdf, const Intersection &isect,
        const Vector &w, float maxDistSquared, bool includeDiffuse);
    // Return indirect irradiance using photons
    Spectrum EPhotonIndirect(const Scene *scene, 
                               const Point &p, 
                               const Normal &n,
                               int numSamples) const;
    
    // SubsurfaceIntegrator Private Data
    u_int nCausticPhotons, nIndirectPhotons, nDirectPhotons;
    u_int nLookup;
    mutable int specularDepth;
    int maxSpecularDepth, maxIndirectDepth;
    float maxDistSquared, maxError;
    bool finalGather, useIrradianceCache;
    int gatherSamples, nSamples, irradianceSamples;
    // Declare sample parameters for light source sampling
    int *lightSampleOffset, lightNumOffset;
    int *bsdfSampleOffset, *bsdfComponentOffset;
    int gatherSampleOffset, gatherComponentOffset;
    int nCausticPaths, nDirectPaths, nIndirectPaths;
    // Photon maps
    mutable vector<Photon> causticPhotons;
    mutable vector<Photon> directPhotons;
    mutable vector<Photon> indirectPhotons;
    mutable KdTree<Photon, PhotonProcess> *causticMap;
    mutable KdTree<Photon, PhotonProcess> *directMap;
    mutable KdTree<Photon, PhotonProcess> *indirectMap;
    mutable Octree<IrradianceSample, IrradProcess> *irradianceCache;
};

struct Photon {
    // Photon Constructor
    Photon(const Point &pp, const Spectrum &wt, const Vector &w)
        : p(pp), alpha(wt), wi(w) {
    }
    Photon(const Point &pp, const Normal &nn, const Spectrum &wt, const Vector &w)
        : p(pp), n(nn), alpha(wt), wi(w) {
    }
    Photon() { }
    Point p;
    Normal n;
    Spectrum alpha;
    Vector wi;
};

inline ostream& operator<<(ostream& os, const Photon &p) {
    os << "Photon [ " << ", p: " << p.p << ", n: " << p.n << ", alpha: " <<
        p.alpha << ", wi: " << p.wi << "]" << endl;
    return os;
}

struct PhotonProcess {
    // PhotonProcess Public Methods
    PhotonProcess(u_int mp, const Point &p);
    void operator()(const Photon &photon, float dist2, float &maxDistSquared) const;
    const Point &p;
    ClosePhoton *photons;
    u_int nLookup;
    mutable u_int foundPhotons;
};

struct ClosePhoton {
    ClosePhoton(const Photon *p = NULL,
                float md2 = INFINITY) {
        photon = p;
        distanceSquared = md2;
    }
    bool operator<(const ClosePhoton &p2) const {
        return distanceSquared == p2.distanceSquared ? (photon < p2.photon) :
            distanceSquared < p2.distanceSquared;
    }
    const Photon *photon;
    float distanceSquared;
};

struct IrradianceSample {
    // IrradianceSample Constructor
    IrradianceSample() { }
    IrradianceSample(const Spectrum &e, const Point &P, const Normal &N,
        float md) : E(e), n(N), p(P) {
        maxDist = md;
    }
    Spectrum E;
    Normal n;
    Point p;
    float maxDist;
};

struct IrradProcess {
    // IrradProcess Public Methods
    IrradProcess(const Normal &N, float me) {
        n = N;
        maxError = me;
        nFound = samplesChecked = 0;
        sumWt = 0.;
        E = 0.;
    }
    void operator()(const Point &P, const IrradianceSample &sample) const;
    bool Successful() {
        return (sumWt > 0. && nFound > 0);
    }
    Spectrum GetIrradiance() const { return E / sumWt; }
    Normal n;
    float maxError;
    mutable int nFound, samplesChecked;
    mutable float sumWt;
    mutable Spectrum E;
};

// Photonmap Method Definitions
SubsurfaceIntegrator::SubsurfaceIntegrator(int ncaus, int ndir, int nind,
        int nl, int ns, int mdepth, float mdist, bool fg, int gs, int is, float me, bool ic) {
    nCausticPhotons = ncaus;
    nIndirectPhotons = nind;
    nDirectPhotons = ndir;
    nLookup = nl;
    nSamples = ns;
    maxDistSquared = mdist * mdist;
    maxSpecularDepth = mdepth;
    maxIndirectDepth = mdepth;
    causticMap = directMap = indirectMap = NULL;
    specularDepth = 0;
    finalGather = fg;
    gatherSamples = gs;
    irradianceSamples = is;
    maxError = me;
    useIrradianceCache = ic;
}

SubsurfaceIntegrator::~SubsurfaceIntegrator() {
    delete causticMap;
    delete directMap;
    delete indirectMap;
    delete irradianceCache;
}

void SubsurfaceIntegrator::RequestSamples(Sample *sample,
        const Scene *scene) {
    // Allocate and request samples for sampling all lights
    u_int nLights = scene->lights.size();
    lightSampleOffset = new int[nLights];
    bsdfSampleOffset = new int[nLights];
    bsdfComponentOffset = new int[nLights];
    for (u_int i = 0; i < nLights; ++i) {
        const Light *light = scene->lights[i];
        int lightSamples = scene->sampler->RoundSize(light->nSamples);
        lightSampleOffset[i] = sample->Add2D(lightSamples);
        bsdfSampleOffset[i] = sample->Add2D(lightSamples);
        bsdfComponentOffset[i] = sample->Add1D(lightSamples);
    }
    lightNumOffset = -1;
    if (finalGather) {
        gatherSamples = scene->sampler->RoundSize(gatherSamples);
        gatherSampleOffset = sample->Add2D(gatherSamples);
        gatherComponentOffset = sample->Add1D(gatherSamples);
    }
}

void SubsurfaceIntegrator::Preprocess(const Scene *scene) {
    if (scene->lights.size() == 0) return;
    // shoot photons and initialize photon maps
    initPhotonMaps(scene);
    // initialize irradiance cache
    BBox wb = scene->WorldBound();
    Vector delta = .01f * (wb.pMax - wb.pMin);
    wb.pMin -= delta;
    wb.pMax += delta;
    irradianceCache = new Octree<IrradianceSample, IrradProcess>(wb, 8);
    // initialize photon irradiance values
    //initPhotonIrradiance(scene);
    // compute samples for subsurface objects
    computeSubsurfaceSamples(scene);
}

void SubsurfaceIntegrator::initPhotonMaps(const Scene *scene) {
    // Initialize photon shooting statistics
    static StatsCounter nshot("Subsurface Photon Map", "Number of photons shot from lights");
    bool causticDone = (nCausticPhotons == 0);
    bool directDone = (nDirectPhotons == 0);
    bool indirectDone = (nIndirectPhotons == 0);
    ProgressReporter progress(nCausticPhotons + nDirectPhotons + nIndirectPhotons, "Shooting photons");
    
    // Shoot photons
    while (!causticDone || !directDone || !indirectDone) {
        ++nshot;
        // Give up if we're not storing enough photons
        if (nshot > 500000 &&
            (unsuccessful(nCausticPhotons, causticPhotons.size(), nshot) ||
             unsuccessful(nDirectPhotons, directPhotons.size(), nshot) ||
             unsuccessful(nIndirectPhotons, indirectPhotons.size(), nshot))) {
            Error("Unable to store enough photons.  Giving up.\n");
            return;
        }
        
        // Choose 4D sample values for photon
        float u[4];
        u[0] = (float)RadicalInverse((int)nshot+1, 2);
        u[1] = (float)RadicalInverse((int)nshot+1, 3);
        u[2] = (float)RadicalInverse((int)nshot+1, 5);
        u[3] = (float)RadicalInverse((int)nshot+1, 7);
        
        // Choose light to shoot photon from
        int nLights = int(scene->lights.size());
        int lightNum = min(Floor2Int(nLights * (float)RadicalInverse((int)nshot+1, 11)), nLights-1);
        Light *light = scene->lights[lightNum];
        float lightPdf = 1.f / nLights;
        
        // Generate _photonRay_ from light source and initialize _alpha_
        RayDifferential photonRay;
        float pdf;
        Spectrum alpha = light->Sample_L(scene, u[0], u[1], u[2], u[3], &photonRay, &pdf);
        if (pdf == 0.f || alpha.Black()) continue;
        alpha /= pdf * lightPdf;
        if (!alpha.Black()) {
            // Trace photon paths and store contribution in photon maps
            tracePhotons(scene, photonRay, alpha, &causticDone, &directDone, &indirectDone, nshot, progress);
        }
        BSDF::FreeAll();
    }
    progress.Done();
}

void SubsurfaceIntegrator::tracePhotons(const Scene *scene, RayDifferential &ray,
        Spectrum &alpha, bool *causticDone, bool *directDone, bool *indirectDone,
        StatsCounter &nshot, ProgressReporter &progress) {
    // Reserve photon map space
    causticPhotons.reserve(nCausticPhotons);
    directPhotons.reserve(nDirectPhotons);
    indirectPhotons.reserve(nIndirectPhotons);
    
    // Follow photon path through scene and record intersections
    bool specularPath = false;
    Intersection photonIsect;
    int nIntersections = 0;
    while (scene->Intersect(ray, &photonIsect)) {
        ++nIntersections;
        // Handle photon/surface intersection
        alpha *= scene->Transmittance(ray);
        Vector wo = -ray.d;
        BSDF *photonBSDF = photonIsect.GetBSDF(ray);
        BxDFType specularType = BxDFType(BSDF_REFLECTION | BSDF_TRANSMISSION | BSDF_SPECULAR);
        bool hasNonSpecular = (photonBSDF->NumComponents() > photonBSDF->NumComponents(specularType));
        if (hasNonSpecular) {
            // Deposit photon at surface
            Photon photon(photonIsect.dg.p, photonIsect.dg.nn, alpha, wo);
            if (nIntersections == 1) {
                // Process direct lighting photon intersection
                if (!*directDone) {
                    directPhotons.push_back(photon);
                    if (directPhotons.size() == nDirectPhotons) {
                        *directDone = true;
                        nDirectPaths = (int)nshot;
                        directMap = new KdTree<Photon, PhotonProcess>(directPhotons);
                    }
                    progress.Update();
                }
            }
            else if (specularPath) {
                // Process caustic photon intersection
                if (!*causticDone) {
                    causticPhotons.push_back(photon);
                    if (causticPhotons.size() == nCausticPhotons) {
                        *causticDone = true;
                        nCausticPaths = (int)nshot;
                        causticMap = new KdTree<Photon, PhotonProcess>(causticPhotons);
                    }
                    progress.Update();
                }
            }
            else {
                // Process indirect lighting photon intersection
                if (!*indirectDone) {
                    indirectPhotons.push_back(photon);
                    if (indirectPhotons.size() == nIndirectPhotons) {
                        *indirectDone = true;
                        nIndirectPaths = (int)nshot;
                        indirectMap = new KdTree<Photon, PhotonProcess>(indirectPhotons);
                    }
                    progress.Update();
                }
            }
        }
        
        // Sample new photon ray direction
        Vector wi;
        float pdf;
        BxDFType flags;
        
        // Get random numbers for sampling outgoing photon direction
        float u1, u2, u3;
        if (nIntersections == 1) {
            u1 = (float)RadicalInverse((int)nshot+1, 13);
            u2 = (float)RadicalInverse((int)nshot+1, 17);
            u3 = (float)RadicalInverse((int)nshot+1, 19);
        }
        else {
            u1 = RandomFloat();
            u2 = RandomFloat();
            u3 = RandomFloat();
        }
        Spectrum fr = photonBSDF->Sample_f(wo, &wi, u1, u2, u3, &pdf, BSDF_ALL, &flags);
        if (fr.Black() || pdf == 0.f)
            break;
        specularPath = (nIntersections == 1 || specularPath) && ((flags & BSDF_SPECULAR) != 0);
        alpha *= fr * AbsDot(wi, photonBSDF->dgShading.nn) / pdf;
        ray = RayDifferential(photonIsect.dg.p, wi);
        
        // Possibly terminate photon path
        if (nIntersections > 3) {
            float continueProbability = .5f;
            if (RandomFloat() > continueProbability)
                break;
            alpha /= continueProbability;
        }
    }
}

void SubsurfaceIntegrator::initPhotonIrradiance(const Scene *scene) {
    ProgressReporter progress(causticPhotons.size() + directPhotons.size() +
        indirectPhotons.size(), "Computing irradiance");
    
    // compute irradiance for all photon types
    for (u_int i = 0; i < causticPhotons.size(); i++) {
        addIrradianceSample(scene, computeIrradiance(scene, causticPhotons[i].p,
            causticPhotons[i].n), causticPhotons[i].p, causticPhotons[i].n, maxDistSquared);
        progress.Update();
    }
    for (u_int i = 0; i < directPhotons.size(); i++) {
        addIrradianceSample(scene, computeIrradiance(scene, directPhotons[i].p,
            directPhotons[i].n), directPhotons[i].p, directPhotons[i].n, maxDistSquared);
        progress.Update();
    }
    for (u_int i = 0; i < indirectPhotons.size(); i++) {
        addIrradianceSample(scene, computeIrradiance(scene, indirectPhotons[i].p,
            indirectPhotons[i].n), indirectPhotons[i].p, indirectPhotons[i].n, maxDistSquared);
        progress.Update();
    }
    
    progress.Done();
}

Spectrum SubsurfaceIntegrator::Li(const Scene *scene, const RayDifferential &ray,
        const Sample *sample, float *alpha) const {
    // Compute reflected radiance with photon map
    Spectrum L(0.);
    Intersection isect;
    if (scene->Intersect(ray, &isect)) {
        if (alpha) *alpha = 1.;
        Vector wo = -ray.d;
        
        // Compute emitted light if ray hit an area light source
        L += isect.Le(wo);
        
        // Evaluate BSDF at hit point
        BSDF *bsdf = isect.GetBSDF(ray);
        const Point &p = bsdf->dgShading.p;
        const Normal &n = bsdf->dgShading.nn;
        
        if (isect.primitive->subSurfaceId != 0) {
            // Compute subsurface scattering
            SubSurface& ss = getSubSurfaceMap()[isect.primitive->subSurfaceId];
            L += ss.Lo(p, n, wo);
            //L += computeIrradiance(scene, p, n) * (Spectrum(1.0) / M_PI);
            
            // Compute indirect lighting without diffuse term
            if (finalGather) {
                Spectrum Li(0.);
                // Do one-bounce final gather for photon map
                for (int i = 0; i < gatherSamples; ++i) {
                    // Sample random direction for final gather ray
                    Vector wi;
                    float u1 = sample->twoD[gatherSampleOffset][2*i];
                    float u2 = sample->twoD[gatherSampleOffset][2*i+1];
                    float u3 = sample->oneD[gatherComponentOffset][i];
                    float pdf;
                    Spectrum fr = bsdf->Sample_f(wo, &wi, u1, u2, u3,
                        &pdf, BxDFType(BSDF_ALL & ~BSDF_DIFFUSE & ~BSDF_SPECULAR));
                    if (fr.Black() || pdf == 0.f) continue;
                    RayDifferential bounceRay(p, wi);
                    static StatsCounter gatherRays("Subsurface Photon Map", "Final gather rays traced");
                    ++gatherRays;
                    Intersection gatherIsect;
                    if (scene->Intersect(bounceRay, &gatherIsect)) {
                        // Compute exitant radiance at final gather intersection
                        BSDF *gatherBSDF = gatherIsect.GetBSDF(bounceRay);
                        Vector bounceWo = -bounceRay.d;
                        if (gatherIsect.primitive->subSurfaceId != 0) {
                            SubSurface& so = getSubSurfaceMap()[gatherIsect.primitive->subSurfaceId];
                            Li += fr * so.Lo(p, n, bounceWo) * AbsDot(wi, n) / pdf;
                        } else {
                            Spectrum Lindir =
                                LPhoton(directMap, nDirectPaths, nLookup,
                                    gatherBSDF, gatherIsect, bounceWo, maxDistSquared, false) +
                                LPhoton(indirectMap, nIndirectPaths, nLookup,
                                    gatherBSDF, gatherIsect, bounceWo, maxDistSquared, false) +
                                LPhoton(causticMap, nCausticPaths, nLookup,
                                    gatherBSDF, gatherIsect, bounceWo, maxDistSquared, false);
                            Lindir *= scene->Transmittance(bounceRay);
                            Li += fr * Lindir * AbsDot(wi, n) / pdf;
                        }
                    }
                }
                L += Li / float(gatherSamples);
            }
            else {
                L += LPhoton(directMap, nDirectPaths, nLookup,
                    bsdf, isect,  wo, maxDistSquared, false);
                L += LPhoton(indirectMap, nIndirectPaths, nLookup,
                    bsdf, isect, wo, maxDistSquared, false);
            }
            
            // Compute caustic lighting
            L += LPhoton(causticMap, nCausticPaths, nLookup, bsdf, isect, wo, maxDistSquared, false);
        } else {
            // Compute direct lighting with diffuse term
            L += UniformSampleAllLights(scene, p, n, wo, bsdf, sample,
                lightSampleOffset, bsdfSampleOffset, bsdfComponentOffset, true);
            
            // Compute indirect lighting
            if (finalGather) {
                Spectrum Li(0.);
                // Do one-bounce final gather for photon map
                for (int i = 0; i < gatherSamples; ++i) {
                    // Sample random direction for final gather ray
                    Vector wi;
                    float u1 = sample->twoD[gatherSampleOffset][2*i];
                    float u2 = sample->twoD[gatherSampleOffset][2*i+1];
                    float u3 = sample->oneD[gatherComponentOffset][i];
                    float pdf;
                    Spectrum fr = bsdf->Sample_f(wo, &wi, u1, u2, u3,
                        &pdf, BxDFType(BSDF_ALL & ~BSDF_SPECULAR));
                    if (fr.Black() || pdf == 0.f) continue;
                    RayDifferential bounceRay(p, wi);
                    static StatsCounter gatherRays("Subsurface Photon Map", "Final gather rays traced");
                    ++gatherRays;
                    Intersection gatherIsect;
                    if (scene->Intersect(bounceRay, &gatherIsect)) {
                        // Compute exitant radiance at final gather intersection
                        BSDF *gatherBSDF = gatherIsect.GetBSDF(bounceRay);
                        Vector bounceWo = -bounceRay.d;
                        if (gatherIsect.primitive->subSurfaceId != 0) {
                            SubSurface& so = getSubSurfaceMap()[gatherIsect.primitive->subSurfaceId];
                            Li += fr * so.Lo(p, n, bounceWo) * AbsDot(wi, n) / pdf;
                        } else {
                            Spectrum Lindir =
                                LPhoton(directMap, nDirectPaths, nLookup,
                                    gatherBSDF, gatherIsect, bounceWo, maxDistSquared, true) +
                                LPhoton(indirectMap, nIndirectPaths, nLookup,
                                    gatherBSDF, gatherIsect, bounceWo, maxDistSquared, true) +
                                LPhoton(causticMap, nCausticPaths, nLookup,
                                    gatherBSDF, gatherIsect, bounceWo, maxDistSquared, true);
                            Lindir *= scene->Transmittance(bounceRay);
                            Li += fr * Lindir * AbsDot(wi, n) / pdf;
                        }
                    }
                }
                L += Li / float(gatherSamples);
            }
            else {
                L += LPhoton(indirectMap, nIndirectPaths, nLookup,
                    bsdf, isect, wo, maxDistSquared, true);
            }
            
            // Compute caustic lighting
            L += LPhoton(causticMap, nCausticPaths, nLookup, bsdf, isect, wo, maxDistSquared, true);
        }
        
        // Compute specular lighting
        if (specularDepth++ < maxSpecularDepth) {
            Vector wi;
            // Trace rays for specular reflection and refraction
            Spectrum f = bsdf->Sample_f(wo, &wi,
                BxDFType(BSDF_REFLECTION | BSDF_SPECULAR));
            if (!f.Black()) {
                // Compute ray differential _rd_ for specular reflection
                RayDifferential rd(p, wi);
                rd.hasDifferentials = true;
                rd.rx.o = p + isect.dg.dpdx;
                rd.ry.o = p + isect.dg.dpdy;
                // Compute differential reflected directions
                Normal dndx = bsdf->dgShading.dndu * bsdf->dgShading.dudx +
                    bsdf->dgShading.dndv * bsdf->dgShading.dvdx;
                Normal dndy = bsdf->dgShading.dndu * bsdf->dgShading.dudy +
                    bsdf->dgShading.dndv * bsdf->dgShading.dvdy;
                Vector dwodx = -ray.rx.d - wo, dwody = -ray.ry.d - wo;
                float dDNdx = Dot(dwodx, n) + Dot(wo, dndx);
                float dDNdy = Dot(dwody, n) + Dot(wo, dndy);
                rd.rx.d = wi - dwodx + 2 * Vector(Dot(wo, n) * dndx + dDNdx * n);
                rd.ry.d = wi - dwody + 2 * Vector(Dot(wo, n) * dndy + dDNdy * n);
                L += scene->Li(rd, sample) * f * AbsDot(wi, n);
            }
            f = bsdf->Sample_f(wo, &wi,
                BxDFType(BSDF_TRANSMISSION | BSDF_SPECULAR));
            if (!f.Black()) {
                // Compute ray differential _rd_ for specular transmission
                RayDifferential rd(p, wi);
                rd.hasDifferentials = true;
                rd.rx.o = p + isect.dg.dpdx;
                rd.ry.o = p + isect.dg.dpdy;
                
                float eta = bsdf->eta;
                Vector w = -wo;
                if (Dot(wo, n) < 0) eta = 1.f / eta;
                
                Normal dndx = bsdf->dgShading.dndu * bsdf->dgShading.dudx + bsdf->dgShading.dndv * bsdf->dgShading.dvdx;
                Normal dndy = bsdf->dgShading.dndu * bsdf->dgShading.dudy + bsdf->dgShading.dndv * bsdf->dgShading.dvdy;
                
                Vector dwodx = -ray.rx.d - wo, dwody = -ray.ry.d - wo;
                float dDNdx = Dot(dwodx, n) + Dot(wo, dndx);
                float dDNdy = Dot(dwody, n) + Dot(wo, dndy);
                
                float mu = eta * Dot(w, n) - Dot(wi, n);
                float dmudx = (eta - (eta*eta*Dot(w,n))/Dot(wi, n)) * dDNdx;
                float dmudy = (eta - (eta*eta*Dot(w,n))/Dot(wi, n)) * dDNdy;
                
                rd.rx.d = wi + eta * dwodx - Vector(mu * dndx + dmudx * n);
                rd.ry.d = wi + eta * dwody - Vector(mu * dndy + dmudy * n);
                L += scene->Li(rd, sample) * f * AbsDot(wi, n);
            }
        }
        --specularDepth;
    }
    else {
        // Handle ray with no intersection
        if (alpha) *alpha = 0.;
        for (u_int i = 0; i < scene->lights.size(); ++i)
            L += scene->lights[i]->Le(ray);
        if (alpha && !L.Black()) *alpha = 1.;
        return L;
    }
    return L;
}

Spectrum SubsurfaceIntegrator::computeIrradiance(const Scene *scene, const Point &p,
        const Normal &n) const {
    Spectrum E(0.);
    //if (!interpolateIrradiance(scene, p, n, &E)) {
        // Compute irradiance at current point
        
    E += computePhotonIrradiance(scene, p, n);
        
        // Add computed irradiance value to cache
    //    addIrradianceSample(scene, E, p, n, maxDistSquared);
        //cout << "Calculated E: " << E << ", at point " << p << endl;
    //} else {
        //cout << "Interpolated E: " << E << ", at point " << p << endl;
    //}
    return E;
}

bool SubsurfaceIntegrator::interpolateIrradiance(const Scene *scene,
        const Point &p, const Normal &n, Spectrum *E) const {
    if (!irradianceCache) return false;
    IrradProcess proc(n, maxError);
    irradianceCache->Lookup(p, proc);
    // Update irradiance cache lookup statistics
    static StatsPercentage nSuccessfulLookups("Subsurface Irradiance Cache",
        "Successful irradiance cache lookups");
    static StatsRatio nSamplesFound("Subsurface Irradiance Cache",
        "Irradiance samples found per successful lookup");
    static StatsRatio nSamplesChecked("Subsurface Irradiance Cache",
        "Irradiance samples checked per lookup");
    nSuccessfulLookups.Add(proc.Successful() ? 1 : 0, 1);
    nSamplesFound.Add(proc.nFound, 1);
    nSamplesChecked.Add(proc.samplesChecked, 1);
    if (!proc.Successful()) return false;
    *E = proc.GetIrradiance();
    return true;
}

void SubsurfaceIntegrator::addIrradianceSample(const Scene *scene, const Spectrum &E,
    const Point &p, const Normal &n, const float dist) const {
    // Update statistics for new irradiance sample
    static StatsCounter nSamplesComputed("Subsurface Irradiance Cache",
        "Irradiance estimates computed");
    ++nSamplesComputed;
    // Compute bounding box of irradiance sample's contribution region
    static float minMaxDist = .001f * powf(scene->WorldBound().Volume(), 1.f/3.f);
    static float maxMaxDist = .125f * powf(scene->WorldBound().Volume(), 1.f/3.f);
    float maxDist = nSamples / dist;
    if (minMaxDist > 0.f) {
        maxDist = Clamp(maxDist, minMaxDist, maxMaxDist);
    }
    maxDist *= maxError;
    BBox sampleExtent(p);
    sampleExtent.Expand(maxDist);
    irradianceCache->Add(IrradianceSample(E, p, n, maxDist), sampleExtent);   
}

Spectrum SubsurfaceIntegrator::computePhotonIrradiance(const Scene *scene,
        const Point &p, const Normal &n) const {
    Spectrum E(0.);
    
    // use parent to sample lights
    E += computeDirectIrradiance(scene, p, n, irradianceSamples);
    E += EPhotonIndirect(scene, p, n, irradianceSamples);
    
    return E;
}

Spectrum SubsurfaceIntegrator::LPhoton(KdTree<Photon, PhotonProcess> *map,
        int nPaths, int nLookup, BSDF *bsdf, const Intersection &isect,
        const Vector &wo, float maxDistSquared, bool includeDiffuse) {
    Spectrum L(0.);
    if (!map) return L;
    BxDFType nonSpecular = BxDFType(BSDF_REFLECTION | BSDF_TRANSMISSION | BSDF_DIFFUSE | BSDF_GLOSSY);
    if (bsdf->NumComponents(nonSpecular) == 0)
        return L;
    static StatsCounter lookups("Subsurface Photon Map", "Total lookups");
    
    // Initialize _PhotonProcess_ object, _proc_, for photon map lookups
    PhotonProcess proc(nLookup, isect.dg.p);
    proc.photons = (ClosePhoton *)alloca(nLookup * sizeof(ClosePhoton));
    
    // Do photon map lookup
    ++lookups;
    map->Lookup(isect.dg.p, proc, maxDistSquared);
    
    // Accumulate light from nearby photons
    static StatsRatio foundRate("Subsurface Photon Map", "Photons found per lookup");
    foundRate.Add(proc.foundPhotons, 1);
    float scale = 1.f / (float(nPaths) * maxDistSquared * M_PI);
    
    // Estimate reflected light from photons
    ClosePhoton *photons = proc.photons;
    int nFound = proc.foundPhotons;
    Normal Nf = Dot(wo, bsdf->dgShading.nn) < 0 ? -bsdf->dgShading.nn : bsdf->dgShading.nn;
    if (bsdf->NumComponents(BxDFType(BSDF_REFLECTION | BSDF_TRANSMISSION | BSDF_GLOSSY)) > 0) {
        // Compute exitant radiance from photons for glossy surface
        for (int i = 0; i < nFound; ++i) {
            BxDFType flag = Dot(Nf, photons[i].photon->wi) > 0.f ?
                (includeDiffuse ? BSDF_ALL_REFLECTION : BxDFType(BSDF_ALL_REFLECTION & ~BSDF_DIFFUSE)) :
                (includeDiffuse ? BSDF_ALL_TRANSMISSION : BxDFType(BSDF_ALL_TRANSMISSION & ~BSDF_DIFFUSE));
            L += bsdf->f(wo, photons[i].photon->wi, flag) * (scale * photons[i].photon->alpha);
        }
    }
    else {
        // Compute exitant radiance from photons for diffuse surface
        Spectrum Lr(0.), Lt(0.);
        for (int i = 0; i < nFound; ++i)
            if (Dot(Nf, photons[i].photon->wi) > 0.f)
                Lr += photons[i].photon->alpha;
            else
                Lt += photons[i].photon->alpha;
        BxDFType reflectionFlag = includeDiffuse ?
            BSDF_ALL_REFLECTION : BxDFType(BSDF_ALL_REFLECTION & ~BSDF_DIFFUSE);
        BxDFType transmissionFlag = includeDiffuse ?
            BSDF_ALL_TRANSMISSION : BxDFType(BSDF_ALL_TRANSMISSION & ~BSDF_DIFFUSE);
        L += (scale * INV_PI) * (Lr * bsdf->rho(wo, reflectionFlag) +
            Lt * bsdf->rho(wo, transmissionFlag));
    }
    return L;
}

Spectrum SubsurfaceIntegrator::EPhotonIndirect(const Scene *scene, 
                                               const Point &p, 
                                               const Normal &n,
                                               int numSamples) const {
    Spectrum E(0.);
    
    // Compute irradiance at current point
    u_int scramble[2] = { RandomUInt(), RandomUInt() };
    for (int i = 0; i < numSamples; ++i) {
        // Sample random direction for final gather ray
        float u[2];
        Sample02(i, scramble, u);
        Vector w = CosineSampleHemisphere(u[0], u[1]);
        RayDifferential r(p, w);
        if (Dot(w, n) < 0) w = -w;
        Spectrum L(0.);
        Intersection isect;
        if (scene->Intersect(r, &isect)) {
            // Compute exitant radiance at final gather intersection
            BSDF *bsdf = isect.GetBSDF(r);
            Vector bounceWo = -r.d;
            //L += isect.Le(bounceWo);
            
            L += LPhoton(indirectMap, nIndirectPaths, nLookup,
                    bsdf, isect, bounceWo, maxDistSquared, true);
                    
            L += LPhoton(causticMap, nCausticPaths, nLookup,
                    bsdf, isect, bounceWo, maxDistSquared, true);
        }
        E += L;
    }
    E *= M_PI / float(numSamples);
    return E;
}

PhotonProcess::PhotonProcess(u_int mp, const Point &P) : p(P) {
    photons = 0;
    nLookup = mp;
    foundPhotons = 0;
}

void PhotonProcess::operator()(const Photon &photon,
        float distSquared, float &maxDistSquared) const {
    static StatsPercentage discarded("Subsurface Photon Map", "Discarded photons");
    discarded.Add(0, 1);
    if (foundPhotons < nLookup) {
        // Add photon to unordered array of photons
        photons[foundPhotons++] = ClosePhoton(&photon, distSquared);
        if (foundPhotons == nLookup) {
            std::make_heap(&photons[0], &photons[nLookup]);
            maxDistSquared = photons[0].distanceSquared;
        }
    }
    else {
        // Remove most distant photon from heap and add new photon
        discarded.Add(1, 0);
        std::pop_heap(&photons[0], &photons[nLookup]);
        photons[nLookup-1] = ClosePhoton(&photon, distSquared);
        std::push_heap(&photons[0], &photons[nLookup]);
        maxDistSquared = photons[0].distanceSquared;
    }
}

void IrradProcess::operator()(const Point &p,
        const IrradianceSample &sample) const {
    ++samplesChecked;
    // Skip irradiance sample if surface normals are too different
    if (Dot(n, sample.n) < 0.01f)
        return;
    // Skip irradiance sample if it's too far from the sample point
    float d2 = DistanceSquared(p, sample.p);
    if (d2 > sample.maxDist * sample.maxDist)
        return;
    // Skip irradiance sample if it's in front of point being shaded
    Normal navg = sample.n + n;
    if (Dot(p - sample.p, navg) < -.01f)
        return;
    // Compute estimate error term and possibly use sample
    float err = sqrtf(d2) / (sample.maxDist * Dot(n, sample.n));
    if (err < 1.) {
        ++nFound;
        float wt = (1.f - err) * (1.f - err);
        E += wt * sample.E;
        sumWt += wt;
    }
}

extern "C" DLLEXPORT SurfaceIntegrator *CreateSurfaceIntegrator(const ParamSet &params) {
    int nCaustic = params.FindOneInt("causticphotons", 20000);
    int nDirect = params.FindOneInt("directphotons", 100000);
    int nIndirect = params.FindOneInt("indirectphotons", 100000);
    int nUsed = params.FindOneInt("nused", 50);
    int nSamples = params.FindOneInt("nsamples", 4096);
    int maxDepth = params.FindOneInt("maxdepth", 5);
    bool finalGather = params.FindOneBool("finalgather", true);
    int gatherSamples = params.FindOneInt("finalgathersamples", 32);
    float maxDist = params.FindOneFloat("maxdist", .1f);
    float maxError = params.FindOneFloat("maxerror", .2f);
    int irradianceSamples = params.FindOneInt("irradiancesamples", 16);
    bool useIrradianceCache = params.FindOneBool("useirradiancecache", false);
    return new SubsurfaceIntegrator(nCaustic, nDirect, nIndirect,
        nUsed, nSamples, maxDepth, maxDist, finalGather, gatherSamples,
        irradianceSamples, maxError, useIrradianceCache);
}
