
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// transport.cpp*
#include "transport.h"
#include "scene.h"
#include "subsurface.h"
#include <iostream>
using namespace std;
// Integrator Method Definitions
Integrator::~Integrator() {
}
// SurfaceIntegrator Method Definitions

/** compute subsurface irradiance samples */
void SurfaceIntegrator::computeSubsurfaceSamples(const Scene *scene) {
    // get all subsurface objects
    map<int, SubSurface>& ssmap = getSubSurfaceMap();
    int totalPrims = 0;
    int pCnt = 0;
    for (map<int, SubSurface>::iterator iter = ssmap.begin(); iter != ssmap.end(); iter++) {
        totalPrims += iter->second.getPrims().size();
    }
    ProgressReporter progress(totalPrims, "Computing subsurface samples");
    
    for (map<int, SubSurface>::iterator iter = ssmap.begin(); iter != ssmap.end(); iter++) {
        SubSurface& ss = iter->second;
        // find total number of samples
        vector< Reference<Primitive> > &prims = ss.getPrims();
        for (size_t i = 0; i < prims.size(); i++) {
            pCnt++;
            // find number of samples for each primitive
            Reference<Primitive> *rp = &prims[i];
            Reference<GeometricPrimitive> *rgp = (Reference<GeometricPrimitive>*) rp;
            Reference<Shape> shape = (*rgp)->getShape();
            
            // do separate samples for each channel
            // add uniform irradiance samples
            int minluColor = 0;
            float minlu = ss.lu.c[0];
            for (int color = 0; color < 3 ; color++) {
                // find min lu's color
                if (ss.lu.c[color] < minlu) {
                    minlu = ss.lu.c[color];
                    minluColor = color;
                }
            }
            for (int color = minluColor; color < minluColor + 1 ; color++) {
                float lu = ss.lu.c[color];
                float samples = shape->Area() / (M_PI * lu * lu) * ss.sampleFactor;
                
                // make sure there's at least 1 sample for
                // really small polygons
                if (samples < 1) {
                    samples = 1;   
                }
                
                float sampleArea = shape->Area() / samples;
                int numSamples = Round2Int(sqrt(samples));

                
                float *pointSamples = (float *)AllocAligned(2 * numSamples * numSamples * sizeof(float));
                StratifiedSample2D(pointSamples, numSamples, numSamples, true);
                
                
                // add uniform irradiance samples
                float maxPoff = 0;
                for (int i = 0; i < 2 * numSamples * numSamples; i += 2) {
                    Normal nShape;
                    Point pShape = shape->Sample(pointSamples[i], pointSamples[i + 1], &nShape);
                    
                        
                    Spectrum E = computeIrradiance(scene, pShape, nShape);
                    ss.addSample(E, sampleArea, pShape, color);
                    
                    // Free BSDF memory used in sample calcs
                    BSDF::FreeAll();
                    
//                    // compute bump-mapped normals, 
//                    // use isect of point just off shape
//                    Intersection isect;
//                    Vector dir(nShape);
//                    Point pOrigin = pShape + dir; 
//                    RayDifferential ray(pOrigin, -dir);
//                    if ((*rgp)->Intersect(ray, &isect)) {
//                        BSDF *bsdf = isect.GetBSDF(ray);
//                        const Point &p = bsdf->dgShading.p;
//                        const Normal &n = bsdf->dgShading.nn;
//                        
//                        
//                        Spectrum E = computeIrradiance(scene, p, n);
//                        ss.addSample(E, sampleArea, p, color);
//                        
//                        // Free BSDF memory used in sample calcs
//                        BSDF::FreeAll();
//                    } else {
//                        //debug 
//                        cout << "Not intersecting!" << pShape << "  " << nShape <<endl; 
//                    }
                }
                
                progress.Update();
                FreeAligned(pointSamples);
              
                // only log every 100 0prims
                if ((pCnt % 1000) == 0) {
                    cout << "Added SS samples[" << pCnt << "/" << totalPrims << " ";
                    cout << color << "]: " << numSamples * numSamples << " id: ";
                    cout << (*rgp)->subSurfaceId << " area: " << shape->Area() << " lu: " << lu;
                    cout <<  endl;
//                    cout << " maxPoff: " << maxPoff << endl;
                }
            }
            
        }
        // summarize the samples
        ss.summarize();
    }
    progress.Done();
}

// default impl, only directlighting irradiance
Spectrum SurfaceIntegrator::computeIrradiance(const Scene *scene, const Point &p, const Normal &n) const {
    return computeDirectIrradiance(scene, p, n, 16); // default to 16 samples
}

Spectrum SurfaceIntegrator::computeDirectIrradiance(const Scene *scene, 
                                                    const Point &p, 
                                                    const Normal &n,
                                                    int nSamples) const {
    Spectrum L(0.);
    // Compute direct lighting to estimate irradiance by default
    for (u_int i = 0; i < scene->lights.size(); i++) {
        Light *light = scene->lights[i];
        Spectrum Ld(0.);
        for (int j = 0; j < nSamples; ++j) {
            // Random sample lights for direct lighting estimate
            Vector wi;
            float lightPdf;
            VisibilityTester visibility;
            Spectrum Li = light->Sample_L(p, n, RandomFloat(), RandomFloat(),
                &wi, &lightPdf, &visibility);
            if (lightPdf > 0. && !Li.Black() && visibility.Unoccluded(scene)) {
                // Add light's contribution to irradiance
                Li *= visibility.Transmittance(scene);
                Ld += Li * AbsDot(wi, n) / lightPdf;
            }
        }
        L += Ld / nSamples;
    }
    return L;
}

bool SurfaceIntegrator::isSampleLocation(const Point &p) const {
    // return true if the point is at a sample location
    for (u_int i = 0; i < ssSamples.size(); i++) {
        if (Distance(p, ssSamples[i]) < 1) {
            return true;
        }
    }
    return false;
}
// Integrator Utility Functions
COREDLL Spectrum UniformSampleAllLights(const Scene *scene,
		const Point &p, const Normal &n, const Vector &wo,
		BSDF *bsdf, const Sample *sample,
		int *lightSampleOffset, int *bsdfSampleOffset,
		int *bsdfComponentOffset,
        bool includeDiffuse) {
	Spectrum L(0.);
	for (u_int i = 0; i < scene->lights.size(); ++i) {
		Light *light = scene->lights[i];
		int nSamples = (sample && lightSampleOffset) ?
			sample->n2D[lightSampleOffset[i]] : 1;
		// Estimate direct lighting from _light_ samples
		Spectrum Ld(0.);
		for (int j = 0; j < nSamples; ++j)
			Ld += EstimateDirect(scene, light, p, n, wo, bsdf,
				sample, lightSampleOffset[i], bsdfSampleOffset[i],
				bsdfComponentOffset[i], j, includeDiffuse);
		L += Ld / nSamples;
	}
	return L;
}
COREDLL Spectrum UniformSampleOneLight(const Scene *scene,
		const Point &p, const Normal &n,
		const Vector &wo, BSDF *bsdf, const Sample *sample,
		int lightSampleOffset, int lightNumOffset,
		int bsdfSampleOffset, int bsdfComponentOffset) {
	// Randomly choose a single light to sample, _light_
	int nLights = int(scene->lights.size());
	int lightNum;
	if (lightNumOffset != -1)
		lightNum = Floor2Int(sample->oneD[lightNumOffset][0] *
							 nLights);
	else
		lightNum = Floor2Int(RandomFloat() * nLights);
	lightNum = min(lightNum, nLights-1);
	Light *light = scene->lights[lightNum];
	return (float)nLights *
		EstimateDirect(scene, light, p, n, wo, bsdf, sample,
			lightSampleOffset, bsdfSampleOffset,
			bsdfComponentOffset, 0);
}
COREDLL Spectrum WeightedSampleOneLight(const Scene *scene,
		const Point &p, const Normal &n,
		const Vector &wo, BSDF *bsdf,
		const Sample *sample, int lightSampleOffset,
		int lightNumOffset, int bsdfSampleOffset,
		int bsdfComponentOffset, float *&avgY,
		float *&avgYsample, float *&cdf,
		float &overallAvgY) {
	int nLights = int(scene->lights.size());
	// Initialize _avgY_ array if necessary
	if (!avgY) {
		avgY = new float[nLights];
		avgYsample = new float[nLights];
		cdf = new float[nLights+1];
		for (int i = 0; i < nLights; ++i)
			avgY[i] = avgYsample[i] = 0.;
	}
	Spectrum L(0.);
	if (overallAvgY == 0.) {
		// Sample one light uniformly and initialize luminance arrays
		L = UniformSampleOneLight(scene, p, n,
		    wo, bsdf, sample, lightSampleOffset,
			lightNumOffset, bsdfSampleOffset,
			bsdfComponentOffset);
		float luminance = L.y();
		overallAvgY = luminance;
		for (int i = 0; i < nLights; ++i)
			avgY[i] = luminance;
	}
	else {
		// Choose _light_ according to average reflected luminance
		float c, lightSampleWeight;
		for (int i = 0; i < nLights; ++i)
			avgYsample[i] = max(avgY[i], .1f * overallAvgY);
		ComputeStep1dCDF(avgYsample, nLights, &c, cdf);
		float t = SampleStep1d(avgYsample, cdf, c, nLights,
			sample->oneD[lightNumOffset][0], &lightSampleWeight);
		int lightNum = min(Float2Int(nLights * t), nLights-1);
		Light *light = scene->lights[lightNum];
		L = EstimateDirect(scene, light, p, n, wo, bsdf,
			sample, lightSampleOffset, bsdfSampleOffset,
			bsdfComponentOffset, 0);
		// Update _avgY_ array with reflected radiance due to light
		float luminance = L.y();
		avgY[lightNum] =
			ExponentialAverage(avgY[lightNum], luminance, .99f);
		overallAvgY =
			ExponentialAverage(overallAvgY, luminance, .999f);
		L /= lightSampleWeight;
	}
	return L;
}
Spectrum EstimateDirect(const Scene *scene,
        const Light *light, const Point &p,
		const Normal &n, const Vector &wo,
		BSDF *bsdf, const Sample *sample, int lightSamp,
		int bsdfSamp, int bsdfComponent, u_int sampleNum,
        bool includeDiffuse) {
	Spectrum Ld(0.);
	// Find light and BSDF sample values for direct lighting estimate
	float ls1, ls2, bs1, bs2, bcs;
	if (lightSamp != -1 && bsdfSamp != -1 &&
		sampleNum < sample->n2D[lightSamp] &&
		sampleNum < sample->n2D[bsdfSamp]) {
		ls1 = sample->twoD[lightSamp][2*sampleNum];
		ls2 = sample->twoD[lightSamp][2*sampleNum+1];
		bs1 = sample->twoD[bsdfSamp][2*sampleNum];
		bs2 = sample->twoD[bsdfSamp][2*sampleNum+1];
		bcs = sample->oneD[bsdfComponent][sampleNum];
	}
	else {
		ls1 = RandomFloat();
		ls2 = RandomFloat();
		bs1 = RandomFloat();
		bs2 = RandomFloat();
		bcs = RandomFloat();
	}
	// Sample light source with multiple importance sampling
	Vector wi;
	float lightPdf, bsdfPdf;
	VisibilityTester visibility;
	Spectrum Li = light->Sample_L(p, n,
		ls1, ls2, &wi, &lightPdf, &visibility);
	if (lightPdf > 0. && !Li.Black()) {
        BxDFType flags = BSDF_ALL;
        if (!includeDiffuse) {
            flags = BxDFType(BSDF_ALL & ~BSDF_DIFFUSE);
        }
		Spectrum f = bsdf->f(wo, wi, flags);
		if (!f.Black() && visibility.Unoccluded(scene)) {
			// Add light's contribution to reflected radiance
			Li *= visibility.Transmittance(scene);
			if (light->IsDeltaLight())
				Ld += f * Li * AbsDot(wi, n) / lightPdf;
			else {
				bsdfPdf = bsdf->Pdf(wo, wi, flags);
				float weight = PowerHeuristic(1, lightPdf, 1, bsdfPdf);
				Ld += f * Li * AbsDot(wi, n) * weight / lightPdf;
			}
		}
	}
	// Sample BSDF with multiple importance sampling
	if (!light->IsDeltaLight()) {
		BxDFType flags = BxDFType(BSDF_ALL & ~BSDF_SPECULAR);
        if (!includeDiffuse) {
            flags = BxDFType(flags & ~BSDF_DIFFUSE);
        }
		Spectrum f = bsdf->Sample_f(wo, &wi,
			bs1, bs2, bcs, &bsdfPdf, flags);
		if (!f.Black() && bsdfPdf > 0.) {
			lightPdf = light->Pdf(p, n, wi);
			if (lightPdf > 0.) {
				// Add light contribution from BSDF sampling
				float weight = PowerHeuristic(1, bsdfPdf, 1, lightPdf);
				Intersection lightIsect;
				Spectrum Li(0.f);
				RayDifferential ray(p, wi);
				if (scene->Intersect(ray, &lightIsect)) {
					if (lightIsect.primitive->GetAreaLight() == light)
						Li = lightIsect.Le(-wi);
				}
				else
					Li = light->Le(ray);
				if (!Li.Black()) {
					Li *= scene->Transmittance(ray);
					Ld += f * Li * AbsDot(wi, n) * weight / bsdfPdf;
				}
			}
		}
	}
	return Ld;
}
